@@ -6,7 +6,7 @@ set -e
 export LANG='C.UTF-8'
 
 dpkg-checkbuilddeps -d 'python3-apt, apt-btrfs-snapshot, parsewiki, python3-feedparser,
-			python3-mock, xvfb, gir1.2-gtk-3.0, python3-gi, python3-nose, pycodestyle, python3-distutils-extra, python3-update-manager'
+			python3-mock, xvfb, gir1.2-gtk-3.0, python3-gi, python3-nose, pycodestyle, python3-distutils-extra, python3-update-manager, python3-yaml'
 
 # make sure the DistUpgrade symlink is good LP: #1824430
 SYMLINK=$(readlink -q DistUpgrade/DistUpgrade)
@@ -24,11 +24,18 @@ xvfb-run nosetests3
 rm -f ./tests/data-sources-list-test/apt.log
 rm -f ./tests/data-sources-list-test/Ubuntu.mirrors
 
+# update demotions
+# echo "Running demotions"
+(cd utils && ./demotions.py impish jammy > demoted.cfg)
+
+# need demotions for EoL interim releases in case people forget to upgrade
+# before the next release is EoL
+(cd utils && ./demotions.py groovy jammy > demoted.cfg.groovy)
+(cd utils && ./demotions.py hirsute jammy > demoted.cfg.hirsute)
+
 # when this gets enabled, make sure to add symlink in DistUpgrade
-if [ $(stat -c "%Y" utils/demoted.cfg.bionic) -lt $max_age ]; then
-    # echo "Running lts demotions"
-    (cd utils && ./demotions.py bionic focal > demoted.cfg.bionic)
-fi
+# echo "Running lts demotions"
+(cd utils && ./demotions.py focal jammy > demoted.cfg.focal)
 
 # update apt_btrfs_snapshot.py copy, this needs an installed
 # apt-btrfs-snapshot on the build machine
@@ -83,11 +90,8 @@ DEBRELEASE=$(LC_ALL=C dpkg-parsechangelog | sed -n -e '/^Distribution:/s/^Distri
 
 # cleanup
 rm -rf utils/apt/lists utils/apt/*.bin utils/apt/sources.list
-max_age=$(($(date +%s) - 3600))
-if [ $(stat -c "%Y" data/mirrors.cfg) -lt $max_age ]; then
-    # echo "Running update mirrors"
-    (cd utils && ./update_mirrors.py ../data/mirrors.cfg)
-fi
+# echo "Running update mirrors"
+(cd utils && ./update_mirrors.py ../data/mirrors.cfg)
 
 # update version
 DEBVER=$(LC_ALL=C dpkg-parsechangelog |sed -n -e '/^Version:/s/^Version: //p' | sed s/.*://)
