@@ -10,8 +10,7 @@ import unittest
 import shutil
 import tempfile
 import json
-import logging
-import subprocess
+import yaml
 
 from DistUpgrade.DistUpgradeQuirks import DistUpgradeQuirks
 
@@ -438,354 +437,547 @@ class TestQuirks(unittest.TestCase):
         self.assertFalse(len(mock_question.mock_calls))
         self.assertFalse(len(mock_controller.abort.mock_calls))
 
-    def test_fips_libgcrypt_not_installed(self):
+    @mock.patch('os.path.exists')
+    def test_aufs_fail(self, mock_exists):
+        mock_exists.return_value = True
         mock_controller = mock.Mock()
-        mock_cache = dict([
-            ('libgcrypt20', make_mock_pkg(
-                name="libgcrypt20",
-                is_installed=True,
-            )),
-        ])
-        mock_controller.cache = mock_cache
-        q = DistUpgradeQuirks(mock_controller, mock.Mock())
-        with self.assertLogs(logging.getLogger(), logging.DEBUG) as cm:
-            q._fipsLibgcryptDivert()
-        self.assertTrue(any("System is not fips enabled"
-                            in s for s in cm.output))
 
-    def test_fips_libgcrypt_usrmerged(self):
-        mock_controller = mock.Mock()
-        mock_cache = dict([
-            ('libgcrypt20', make_mock_pkg(
-                name="libgcrypt20",
-                is_installed=True,
-            )),
-            ('libgcrypt20-hmac', make_mock_pkg(
-                name="libgcrypt20-hmac",
-                is_installed=True,
-            )),
-            ('usrmerge', make_mock_pkg(
-                name="usrmerge",
-                is_installed=True,
-            )),
-        ])
-        mock_controller.cache = mock_cache
-        q = DistUpgradeQuirks(mock_controller, mock.Mock())
-        with self.assertLogs(logging.getLogger(), logging.DEBUG) as cm:
-            q._fipsLibgcryptDivert()
-        self.assertTrue(any("System is usrmerged" in s for s in cm.output))
-
-    def test_fips_libgcrypt_fixed(self):
-        mock_controller = mock.Mock()
-        mock_cache = dict([
-            ('libgcrypt20', make_mock_pkg(
-                name="libgcrypt20",
-                is_installed=True,
-            )),
-            ('libgcrypt20-hmac', make_mock_pkg(
-                name="libgcrypt20-hmac",
-                is_installed=True,
-            )),
-            ('usrmerge', make_mock_pkg(
-                name="usrmerge",
-                is_installed=False,
-            )),
-        ])
-        hmac_version = mock.Mock()
-        hmac_version.version = '1.8.5-5ubuntu1.fips.1.7'
-        mock_cache['libgcrypt20-hmac'].versions = [hmac_version]
-        mock_controller.cache = mock_cache
-        q = DistUpgradeQuirks(mock_controller, mock.Mock())
-        with self.assertLogs(logging.getLogger(), logging.DEBUG) as cm:
-            q._fipsLibgcryptDivert()
-        self.assertTrue(any("Fixed libgcrypt20-hmac will be installed"
-                            in s for s in cm.output))
-
-    @mock.patch('subprocess.check_call')
-    def test_fips_libgcrypt_broken(self, mocked_check_call):
-        mock_controller = mock.Mock()
-        mock_cache = dict([
-            ('libgcrypt20', make_mock_pkg(
-                name="libgcrypt20",
-                is_installed=True,
-            )),
-            ('libgcrypt20-hmac', make_mock_pkg(
-                name="libgcrypt20-hmac",
-                is_installed=True,
-            )),
-            ('usrmerge', make_mock_pkg(
-                name="usrmerge",
-                is_installed=False,
-            )),
-        ])
-        hmac_version = mock.Mock()
-        hmac_version.version = '1.8.5-5ubuntu1.fips.1.4'
-        mock_cache['libgcrypt20-hmac'].versions = [hmac_version]
-        mock_controller.cache = mock_cache
-        q = DistUpgradeQuirks(mock_controller, mock.Mock())
-        with self.assertLogs(logging.getLogger(), logging.DEBUG) as cm:
-            q._fipsLibgcryptDivert()
-        self.assertTrue(any("Broken libgcrypt20-hmac will be installed"
-                            in s for s in cm.output))
-        self.assertTrue(mocked_check_call.called)
-
-    @mock.patch('subprocess.check_call', side_effect=Exception())
-    def test_fips_libgcrypt_broken_exception(self, mocked_check_call):
-        mock_controller = mock.Mock()
-        mock_cache = dict([
-            ('libgcrypt20', make_mock_pkg(
-                name="libgcrypt20",
-                is_installed=True,
-            )),
-            ('libgcrypt20-hmac', make_mock_pkg(
-                name="libgcrypt20-hmac",
-                is_installed=True,
-            )),
-            ('usrmerge', make_mock_pkg(
-                name="usrmerge",
-                is_installed=False,
-            )),
-        ])
-        hmac_version = mock.Mock()
-        hmac_version.version = '1.8.5-5ubuntu1.fips.1.4'
-        mock_cache['libgcrypt20-hmac'].versions = [hmac_version]
-        mock_controller.cache = mock_cache
-        q = DistUpgradeQuirks(mock_controller, mock.Mock())
-        with self.assertLogs(logging.getLogger(), logging.DEBUG) as cm:
-            q._fipsLibgcryptDivert()
-        self.assertTrue(any("Broken libgcrypt20-hmac will be installed"
-                            in s for s in cm.output))
-        self.assertTrue(any("Failed to add dpkg-divert for"
-                            in s for s in cm.output))
-        self.assertTrue(mocked_check_call.called)
-
-    def test_fips_libgcrypt_cleanup_not_installed(self):
-        mock_controller = mock.Mock()
-        mock_cache = dict([
-            ('libgcrypt20', make_mock_pkg(
-                name="libgcrypt20",
-                is_installed=True,
-            )),
-        ])
-        mock_controller.cache = mock_cache
-        q = DistUpgradeQuirks(mock_controller, mock.Mock())
-        with self.assertLogs(logging.getLogger(), logging.DEBUG) as cm:
-            q._fipsLibgcryptHmacCleanup()
-        self.assertTrue(any("System is not fips enabled"
-                            in s for s in cm.output))
-
-    def test_fips_libgcrypt_cleanup_usrmerged(self):
-        mock_controller = mock.Mock()
-        mock_cache = dict([
-            ('libgcrypt20', make_mock_pkg(
-                name="libgcrypt20",
-                is_installed=True,
-            )),
-            ('libgcrypt20-hmac', make_mock_pkg(
-                name="libgcrypt20-hmac",
-                is_installed=True,
-            )),
-            ('usrmerge', make_mock_pkg(
-                name="usrmerge",
-                is_installed=True,
-            )),
-        ])
-        mock_controller.cache = mock_cache
-        q = DistUpgradeQuirks(mock_controller, mock.Mock())
-        with self.assertLogs(logging.getLogger(), logging.DEBUG) as cm:
-            q._fipsLibgcryptHmacCleanup()
-        self.assertTrue(any("System is usrmerged" in s for s in cm.output))
-
-    def test_fips_libgcrypt_cleanup_fixed(self):
-        mock_controller = mock.Mock()
-        mock_cache = dict([
-            ('libgcrypt20', make_mock_pkg(
-                name="libgcrypt20",
-                is_installed=True,
-            )),
-            ('libgcrypt20-hmac', make_mock_pkg(
-                name="libgcrypt20-hmac",
-                is_installed=True,
-            )),
-            ('usrmerge', make_mock_pkg(
-                name="usrmerge",
-                is_installed=False,
-            )),
-        ])
-        hmac_version = mock.Mock()
-        hmac_version.version = '1.8.5-5ubuntu1.fips.1.7'
-        mock_cache['libgcrypt20-hmac'].installed = hmac_version
-        mock_controller.cache = mock_cache
-        q = DistUpgradeQuirks(mock_controller, mock.Mock())
-        with self.assertLogs(logging.getLogger(), logging.DEBUG) as cm:
-            q._fipsLibgcryptHmacCleanup()
-        self.assertTrue(any("Fixed libgcrypt20-hmac was installed"
-                            in s for s in cm.output))
-
-    @mock.patch('os.path.exists', side_effect=[False, False])
-    def test_fips_libgcrypt_cleanup_broken_old_missing(self, mocked_exists):
-        mock_controller = mock.Mock()
-        mock_cache = dict([
-            ('libgcrypt20', make_mock_pkg(
-                name="libgcrypt20",
-                is_installed=True,
-            )),
-            ('libgcrypt20-hmac', make_mock_pkg(
-                name="libgcrypt20-hmac",
-                is_installed=True,
-            )),
-            ('usrmerge', make_mock_pkg(
-                name="usrmerge",
-                is_installed=False,
-            )),
-        ])
-        hmac_version = mock.Mock()
-        hmac_version.version = '1.8.5-5ubuntu1.fips.1.4'
-        mock_cache['libgcrypt20-hmac'].installed = hmac_version
-        mock_controller.cache = mock_cache
-        q = DistUpgradeQuirks(mock_controller, mock.Mock())
-        with self.assertLogs(logging.getLogger(), logging.DEBUG) as cm:
-            q._fipsLibgcryptHmacCleanup()
-        self.assertTrue(any("Missing hmac file: "
-                            "/lib/x86_64-linux-gnu/.libgcrypt.so.20.hmac "
-                            in s for s in cm.output))
-        self.assertTrue(mocked_exists.called)
-
-    @mock.patch('os.path.exists', side_effect=[True, False])
-    def test_fips_libgcrypt_cleanup_broken_new_missing(
-            self, mocked_exists):
-        mock_controller = mock.Mock()
-        mock_cache = dict([
-            ('libgcrypt20', make_mock_pkg(
-                name="libgcrypt20",
-                is_installed=True,
-            )),
-            ('libgcrypt20-hmac', make_mock_pkg(
-                name="libgcrypt20-hmac",
-                is_installed=True,
-            )),
-            ('usrmerge', make_mock_pkg(
-                name="usrmerge",
-                is_installed=False,
-            )),
-        ])
-        hmac_version = mock.Mock()
-        hmac_version.version = '1.8.5-5ubuntu1.fips.1.4'
-        mock_cache['libgcrypt20-hmac'].installed = hmac_version
-        mock_controller.cache = mock_cache
-        q = DistUpgradeQuirks(mock_controller, mock.Mock())
-        with self.assertLogs(logging.getLogger(), logging.DEBUG) as cm:
-            q._fipsLibgcryptHmacCleanup()
-        self.assertTrue(any("Missing hmac file: "
-                            "/usr/lib/x86_64-linux-gnu/.libgcrypt.so.20.hmac "
-                            in s for s in cm.output))
-        self.assertTrue(mocked_exists.called)
-
-    @mock.patch('os.unlink')
-    @mock.patch('os.path.exists', side_effect=[True, True])
-    @mock.patch('os.path.islink', side_effect=[False])
-    def test_fips_libgcrypt_cleanup_broken_present(
-            self, mocked_islink, mocked_exists, mocked_unlink):
-        mock_controller = mock.Mock()
-        mock_cache = dict([
-            ('libgcrypt20', make_mock_pkg(
-                name="libgcrypt20",
-                is_installed=True,
-            )),
-            ('libgcrypt20-hmac', make_mock_pkg(
-                name="libgcrypt20-hmac",
-                is_installed=True,
-            )),
-            ('usrmerge', make_mock_pkg(
-                name="usrmerge",
-                is_installed=False,
-            )),
-        ])
-        hmac_version = mock.Mock()
-        hmac_version.version = '1.8.5-5ubuntu1.fips.1.4'
-        mock_cache['libgcrypt20-hmac'].installed = hmac_version
-        mock_controller.cache = mock_cache
         q = DistUpgradeQuirks(mock_controller, mock.Mock())
 
-        with self.assertLogs(logging.getLogger(), logging.DEBUG) as cm:
-            q._fipsLibgcryptHmacCleanup()
-        self.assertTrue(any("Removing old hmac file from Bionic version"
-                            in s for s in cm.output))
-        self.assertTrue(mocked_exists.called)
-        self.assertTrue(mocked_unlink.called)
-        self.assertTrue(mocked_islink.called)
+        q._test_and_fail_on_aufs()
+        self.assertTrue(len(mock_controller.abort.mock_calls))
 
-    @mock.patch('os.path.exists', side_effect=[True, True])
-    @mock.patch('os.path.islink', side_effect=[True])
-    @mock.patch('os.path.realpath',
-                side_effect=['/usr/lib/x86_64-linux-gnu/.libgcrypt.so.20.hmac']
-                )
-    def test_fips_libgcrypt_cleanup_broken_present_identical(
-            self, mocked_realpath, mocked_islink, mocked_exists):
-        mock_controller = mock.Mock()
-        mock_cache = dict([
-            ('libgcrypt20', make_mock_pkg(
-                name="libgcrypt20",
-                is_installed=True,
-            )),
-            ('libgcrypt20-hmac', make_mock_pkg(
-                name="libgcrypt20-hmac",
-                is_installed=True,
-            )),
-            ('usrmerge', make_mock_pkg(
-                name="usrmerge",
-                is_installed=False,
-            )),
-        ])
-        hmac_version = mock.Mock()
-        hmac_version.version = '1.8.5-5ubuntu1.fips.1.4'
-        mock_cache['libgcrypt20-hmac'].installed = hmac_version
-        mock_controller.cache = mock_cache
-        q = DistUpgradeQuirks(mock_controller, mock.Mock())
+    @mock.patch('builtins.open')
+    def test_pam_tally_fail(self, mock_open):
+        with tempfile.TemporaryFile(mode='w+') as f:
+            f.write("auth            sufficient      pam_tally.so\n"
+                    "@include common-auth\n"
+                    "@include common-account\n"
+                    "@include common-session\n")
+            f.seek(0)
+            mock_open.return_value = f
+            mock_controller = mock.Mock()
 
-        with self.assertLogs(logging.getLogger(), logging.DEBUG) as cm:
-            q._fipsLibgcryptHmacCleanup()
-        self.assertTrue(any("hmac files are identical, system is possibly"
-                            in s for s in cm.output))
-        self.assertTrue(mocked_exists.called)
-        self.assertTrue(mocked_islink.called)
-        self.assertTrue(mocked_realpath.called)
+            q = DistUpgradeQuirks(mock_controller, mock.Mock())
 
-    @mock.patch('os.unlink', side_effect=Exception())
-    @mock.patch('os.path.exists', side_effect=[True, True])
-    @mock.patch('os.path.islink', side_effect=[False])
-    def test_fips_libgcrypt_cleanup_broken_present_exception(
-            self, mocked_islink, mocked_exists, mocked_unlink):
-        mock_controller = mock.Mock()
-        mock_cache = dict([
-            ('libgcrypt20', make_mock_pkg(
-                name="libgcrypt20",
-                is_installed=True,
-            )),
-            ('libgcrypt20-hmac', make_mock_pkg(
-                name="libgcrypt20-hmac",
-                is_installed=True,
-            )),
-            ('usrmerge', make_mock_pkg(
-                name="usrmerge",
-                is_installed=False,
-            )),
-        ])
-        hmac_version = mock.Mock()
-        hmac_version.version = '1.8.5-5ubuntu1.fips.1.4'
-        mock_cache['libgcrypt20-hmac'].installed = hmac_version
-        mock_controller.cache = mock_cache
-        q = DistUpgradeQuirks(mock_controller, mock.Mock())
+            q._test_and_fail_on_pam_tally()
+            self.assertTrue(len(mock_controller.abort.mock_calls))
 
-        with self.assertLogs(logging.getLogger(), logging.DEBUG) as cm:
-            q._fipsLibgcryptHmacCleanup()
-        self.assertTrue(any("Removing old hmac file from Bionic version"
-                            in s for s in cm.output))
-        self.assertTrue(any("Failed to remove"
-                            in s for s in cm.output))
-        self.assertTrue(mocked_exists.called)
-        self.assertTrue(mocked_unlink.called)
-        self.assertTrue(mocked_islink.called)
+    def test_replace_fkms_overlay_no_config(self):
+        with tempfile.TemporaryDirectory() as boot_dir:
+            mock_controller = mock.Mock()
+
+            q = DistUpgradeQuirks(mock_controller, mock.Mock())
+
+            q._replace_fkms_overlay(boot_dir)
+            self.assertFalse(os.path.exists(os.path.join(
+                boot_dir, 'config.txt.distUpgrade')))
+
+    def test_replace_fkms_overlay_no_changes(self):
+        with tempfile.TemporaryDirectory() as boot_dir:
+            demo_config = """\
+# This is a demo boot config
+[pi4]
+max_framebuffers=2
+[all]
+arm_64bit=1
+kernel=vmlinuz
+initramfs initrd.img followkernel"""
+            with open(os.path.join(boot_dir, 'config.txt'), 'w') as f:
+                f.write(demo_config)
+
+            mock_controller = mock.Mock()
+
+            q = DistUpgradeQuirks(mock_controller, mock.Mock())
+
+            q._replace_fkms_overlay(boot_dir)
+            self.assertFalse(os.path.exists(os.path.join(
+                boot_dir, 'config.txt.distUpgrade')))
+            with open(os.path.join(boot_dir, 'config.txt')) as f:
+                self.assertTrue(f.read() == demo_config)
+
+    def test_replace_fkms_overlay_with_changes(self):
+        with tempfile.TemporaryDirectory() as boot_dir:
+            demo_config = """\
+# This is a demo boot config
+[pi4]
+max_framebuffers=2
+[all]
+arm_64bit=1
+kernel=vmlinuz
+initramfs initrd.img followkernel
+dtoverlay=vc4-fkms-v3d,cma-256
+start_x=1
+gpu_mem=256
+"""
+            expected_config = """\
+# This is a demo boot config
+[pi4]
+max_framebuffers=2
+[all]
+arm_64bit=1
+kernel=vmlinuz
+initramfs initrd.img followkernel
+# changed by do-release-upgrade (LP: #1923673)
+#dtoverlay=vc4-fkms-v3d,cma-256
+dtoverlay=vc4-kms-v3d,cma-256
+# disabled by do-release-upgrade (LP: #1923673)
+#start_x=1
+# disabled by do-release-upgrade (LP: #1923673)
+#gpu_mem=256
+"""
+            with open(os.path.join(boot_dir, 'config.txt'), 'w') as f:
+                f.write(demo_config)
+
+            mock_controller = mock.Mock()
+
+            q = DistUpgradeQuirks(mock_controller, mock.Mock())
+
+            q._replace_fkms_overlay(boot_dir)
+            self.assertTrue(os.path.exists(os.path.join(
+                boot_dir, 'config.txt.distUpgrade')))
+            with open(os.path.join(boot_dir, 'config.txt')) as f:
+                self.assertTrue(f.read() == expected_config)
+
+    def test_remove_uboot_no_config(self):
+        with tempfile.TemporaryDirectory() as boot_dir:
+            mock_controller = mock.Mock()
+            q = DistUpgradeQuirks(mock_controller, mock.Mock())
+            q._remove_uboot_on_rpi(boot_dir)
+
+            self.assertFalse(os.path.exists(os.path.join(
+                boot_dir, 'config.txt.distUpgrade')))
+
+    def test_remove_uboot_no_changes(self):
+        with tempfile.TemporaryDirectory() as boot_dir:
+            native_config = """\
+# This is a demo boot config with a comment at the start that should not
+# be removed
+
+[pi4]
+max_framebuffers=2
+
+[all]
+arm_64bit=1
+kernel=vmlinuz
+initramfs initrd.img followkernel
+
+# This is a user-added include that should not be merged
+include custom.txt
+"""
+            custom_config = """\
+# This is the custom included configuration file
+
+hdmi_group=1
+hdmi_mode=4
+"""
+            with open(os.path.join(boot_dir, 'config.txt'), 'w') as f:
+                f.write(native_config)
+            with open(os.path.join(boot_dir, 'custom.txt'), 'w') as f:
+                f.write(custom_config)
+
+            mock_controller = mock.Mock()
+            q = DistUpgradeQuirks(mock_controller, mock.Mock())
+            q._remove_uboot_on_rpi(boot_dir)
+
+            self.assertFalse(os.path.exists(os.path.join(
+                boot_dir, 'config.txt.distUpgrade')))
+            self.assertFalse(os.path.exists(os.path.join(
+                boot_dir, 'custom.txt.distUpgrade')))
+            with open(os.path.join(boot_dir, 'config.txt')) as f:
+                self.assertTrue(f.read() == native_config)
+            with open(os.path.join(boot_dir, 'custom.txt')) as f:
+                self.assertTrue(f.read() == custom_config)
+
+    def test_remove_uboot_with_changes(self):
+        with tempfile.TemporaryDirectory() as boot_dir:
+            config_txt = """\
+# This is a warning that you should not edit this file. The upgrade should
+# remove this comment
+
+[pi4]
+# This is a comment that should be included
+kernel=uboot_rpi_4.bin
+max_framebuffers=2
+
+[pi2]
+kernel=uboot_rpi_2.bin
+
+[pi3]
+kernel=uboot_rpi_3.bin
+
+[all]
+arm_64bit=1
+device_tree_address=0x3000000
+include syscfg.txt
+include usercfg.txt
+dtoverlay=vc4-fkms-v3d,cma-256
+include custom.txt
+"""
+            usercfg_txt = """\
+# Another chunk of warning text that should be skipped
+"""
+            syscfg_txt = """\
+# Yet more warnings to exclude
+dtparam=audio=on
+dtparam=spi=on
+enable_uart=1
+"""
+            custom_txt = """\
+# This is a user-added file that should be left alone by the upgrade
+[gpio4=1]
+kernel=custom
+"""
+            expected_config_txt = """\
+
+[pi4]
+# This is a comment that should be included
+# commented by do-release-upgrade (LP: #1936401)
+#kernel=uboot_rpi_4.bin
+max_framebuffers=2
+
+[pi2]
+# commented by do-release-upgrade (LP: #1936401)
+#kernel=uboot_rpi_2.bin
+
+[pi3]
+# commented by do-release-upgrade (LP: #1936401)
+#kernel=uboot_rpi_3.bin
+
+[all]
+# added by do-release-upgrade (LP: #1936401)
+kernel=vmlinuz
+initramfs initrd.img followkernel
+arm_64bit=1
+# commented by do-release-upgrade (LP: #1936401)
+#device_tree_address=0x3000000
+# merged from syscfg.txt by do-release-upgrade (LP: #1936401)
+dtparam=audio=on
+dtparam=spi=on
+enable_uart=1
+# merged from usercfg.txt by do-release-upgrade (LP: #1936401)
+dtoverlay=vc4-fkms-v3d,cma-256
+include custom.txt
+"""
+            with open(os.path.join(boot_dir, 'config.txt'), 'w') as f:
+                f.write(config_txt)
+            with open(os.path.join(boot_dir, 'syscfg.txt'), 'w') as f:
+                f.write(syscfg_txt)
+            with open(os.path.join(boot_dir, 'usercfg.txt'), 'w') as f:
+                f.write(usercfg_txt)
+            with open(os.path.join(boot_dir, 'custom.txt'), 'w') as f:
+                f.write(custom_txt)
+
+            mock_controller = mock.Mock()
+            q = DistUpgradeQuirks(mock_controller, mock.Mock())
+            q._remove_uboot_on_rpi(boot_dir)
+
+            self.assertTrue(os.path.exists(os.path.join(
+                boot_dir, 'config.txt.distUpgrade')))
+            self.assertTrue(os.path.exists(os.path.join(
+                boot_dir, 'syscfg.txt.distUpgrade')))
+            self.assertTrue(os.path.exists(os.path.join(
+                boot_dir, 'usercfg.txt.distUpgrade')))
+            self.assertTrue(os.path.exists(os.path.join(
+                boot_dir, 'config.txt')))
+            self.assertTrue(os.path.exists(os.path.join(
+                boot_dir, 'custom.txt')))
+            self.assertFalse(os.path.exists(os.path.join(
+                boot_dir, 'syscfg.txt')))
+            self.assertFalse(os.path.exists(os.path.join(
+                boot_dir, 'usercfg.txt')))
+            self.assertFalse(os.path.exists(os.path.join(
+                boot_dir, 'custom.txt.distUpgrade')))
+            with open(os.path.join(boot_dir, 'config.txt')) as f:
+                self.assertTrue(f.read() == expected_config_txt)
+            with open(os.path.join(boot_dir, 'custom.txt')) as f:
+                self.assertTrue(f.read() == custom_txt)
+
+    def test_remove_uboot_no_all_section(self):
+        with tempfile.TemporaryDirectory() as boot_dir:
+            config_txt = """\
+arm_64bit=1
+device_tree_address=0x3000000
+
+[pi4]
+# This is a comment that should be included
+kernel=uboot_rpi_4.bin
+max_framebuffers=2
+
+[pi3]
+kernel=uboot_rpi_3.bin
+"""
+            expected_config_txt = """\
+arm_64bit=1
+# commented by do-release-upgrade (LP: #1936401)
+#device_tree_address=0x3000000
+
+[pi4]
+# This is a comment that should be included
+# commented by do-release-upgrade (LP: #1936401)
+#kernel=uboot_rpi_4.bin
+max_framebuffers=2
+
+[pi3]
+# commented by do-release-upgrade (LP: #1936401)
+#kernel=uboot_rpi_3.bin
+# added by do-release-upgrade (LP: #1936401)
+[all]
+kernel=vmlinuz
+initramfs initrd.img followkernel
+"""
+            with open(os.path.join(boot_dir, 'config.txt'), 'w') as f:
+                f.write(config_txt)
+
+            mock_controller = mock.Mock()
+            q = DistUpgradeQuirks(mock_controller, mock.Mock())
+            q._remove_uboot_on_rpi(boot_dir)
+
+            self.assertTrue(os.path.exists(os.path.join(
+                boot_dir, 'config.txt.distUpgrade')))
+            with open(os.path.join(boot_dir, 'config.txt')) as f:
+                self.assertTrue(f.read() == expected_config_txt)
+
+    # This test case tests the scenario where the invalid
+    # match section is present and needs to be removed
+    def test_remove_netplan_match_single_removal(self):
+        with tempfile.TemporaryDirectory() as netplan_dir:
+            config_yaml = """\
+network:
+  ethernets:
+    # this is a test comment to ensure it is preserved in the backup
+    eth0:
+      dhcp4: true
+      # this section will be deleted
+      match:
+        driver: bcmgenet smsc95xx lan78xx
+        name: en*
+      optional: true
+      set-name: eth0
+  wifis:
+    wlan0:
+      dhcp4: true
+      access-points:
+        "TestWifi":
+          password: "TestPassword"
+  version: 2
+"""
+            expected_config_yaml = """\
+network:
+  ethernets:
+    eth0:
+      dhcp4: true
+      optional: true
+  wifis:
+    wlan0:
+      dhcp4: true
+      access-points:
+        "TestWifi":
+          password: "TestPassword"
+  version: 2
+"""
+            config_filename = '99-test-netplan.yaml'
+            with open(os.path.join(netplan_dir, config_filename), 'w') as f:
+                f.write(config_yaml)
+
+            mock_controller = mock.Mock()
+            q = DistUpgradeQuirks(mock_controller, mock.Mock())
+            q._remove_netplan_match_rpi(netplan_dir)
+
+            # make sure the backup exists and is the same as the original
+            backup_config_filename = config_filename + '.distUpgrade'
+            self.assertTrue(os.path.exists(os.path.join(
+                netplan_dir, backup_config_filename)))
+            with open(os.path.join(netplan_dir,
+                      backup_config_filename)) as f:
+                self.assertTrue(f.read() == config_yaml)
+
+            # make sure the new contents are what we expect
+            with open(os.path.join(netplan_dir, config_filename)) as f:
+                new_config = yaml.load(f, yaml.SafeLoader)
+                expected_config = yaml.load(expected_config_yaml,
+                                            yaml.SafeLoader)
+                self.assertTrue(new_config == expected_config)
+
+    # This test case tests the scenario where the match
+    # section is present but is valid syntax
+    def test_remove_netplan_match_valid_syntax(self):
+        with tempfile.TemporaryDirectory() as netplan_dir:
+            mock_controller = mock.Mock()
+            q = DistUpgradeQuirks(mock_controller, mock.Mock())
+            q._remove_netplan_match_rpi(netplan_dir)
+
+            self.assertFalse(os.path.exists(os.path.join(
+                netplan_dir, '50-cloud-init.yaml.distUpgrade')))
+
+    # This test case tests the scenario where the match
+    # section not present at all
+    def test_remove_netplan_match_no_match(self):
+        with tempfile.TemporaryDirectory() as netplan_dir:
+            mock_controller = mock.Mock()
+            q = DistUpgradeQuirks(mock_controller, mock.Mock())
+            q._remove_netplan_match_rpi(netplan_dir)
+
+            self.assertFalse(os.path.exists(os.path.join(
+                netplan_dir, '50-cloud-init.yaml.distUpgrade')))
+
+    # This test case tests the scenario where the invalid
+    # match section is present in one file, but multiple
+    # config files exist
+    def test_remove_netplan_match_one_of_multiple(self):
+        with tempfile.TemporaryDirectory() as netplan_dir:
+            ethernet_config_yaml = """\
+############
+# also
+# test
+# multi
+# line
+# comments
+############
+network:
+  ethernets:
+    eth0:
+      dhcp4: true
+      match:
+        driver: bcmgenet smsc95xx lan78xx
+      optional: true
+      set-name: eth0
+  version: 2
+"""
+            wifi_config_yaml = """\
+network:
+  wifis:
+    wlan0:
+      dhcp4: true
+      access-points:
+        "TestWifi":
+          password: "TestPassword"
+  version: 2
+"""
+            expected_ethernet_config_yaml = """\
+network:
+  ethernets:
+    eth0:
+      dhcp4: true
+      optional: true
+  version: 2
+"""
+            e_config_filename = '01-test-ethernet.yaml'
+            with open(os.path.join(netplan_dir, e_config_filename), 'w') as f:
+                f.write(ethernet_config_yaml)
+
+            w_config_filename = '02-test-wifi.yaml'
+            with open(os.path.join(netplan_dir, w_config_filename), 'w') as f:
+                f.write(wifi_config_yaml)
+
+            mock_controller = mock.Mock()
+            q = DistUpgradeQuirks(mock_controller, mock.Mock())
+            q._remove_netplan_match_rpi(netplan_dir)
+
+            # make sure the backup exists and is the same as the original
+            backup_config_filename = e_config_filename + '.distUpgrade'
+            self.assertTrue(os.path.exists(os.path.join(
+                netplan_dir, backup_config_filename)))
+            with open(os.path.join(netplan_dir,
+                      backup_config_filename)) as f:
+                self.assertTrue(f.read() == ethernet_config_yaml)
+
+            # make sure the new contents are what we expect
+            with open(os.path.join(netplan_dir, e_config_filename)) as f:
+                new_config = yaml.load(f, yaml.SafeLoader)
+                expected_config = yaml.load(expected_ethernet_config_yaml,
+                                            yaml.SafeLoader)
+                self.assertTrue(new_config == expected_config)
+
+            # make sure a backup wasn't created for the config
+            # that doesn't need modification
+            self.assertFalse(os.path.exists(os.path.join(
+                netplan_dir, w_config_filename + '.distUpgrade')))
+
+    # This test case tests the scenario where the invalid
+    # match section is present in multiple files
+    def test_remove_netplan_match_multiple(self):
+        with tempfile.TemporaryDirectory() as netplan_dir:
+            eth0_config_yaml = """\
+# comments can go at the top
+network:
+  ethernets:
+    eth0:
+      dhcp4: true
+      match:
+        driver: bcmgenet smsc95xx lan78xx
+      optional: true
+      set-name: eth0
+  version: 2
+"""
+            eth1_config_yaml = """\
+network:
+  ethernets:
+    eth1:
+      dhcp4: true
+      match:
+        driver: bcmgenet smsc95xx lan78xx
+  version: 2
+# comments can go at the bottom too
+"""
+            expected_eth0_config_yaml = """\
+network:
+  ethernets:
+    eth0:
+      dhcp4: true
+      optional: true
+  version: 2
+"""
+            expected_eth1_config_yaml = """\
+network:
+  ethernets:
+    eth1:
+      dhcp4: true
+  version: 2
+"""
+            eth0_config_filename = '01-test-eth0.yaml'
+            with open(os.path.join(netplan_dir,
+                      eth0_config_filename), 'w') as f:
+                f.write(eth0_config_yaml)
+
+            eth1_config_filename = '02-test-eth1.yaml'
+            with open(os.path.join(netplan_dir,
+                      eth1_config_filename), 'w') as f:
+                f.write(eth1_config_yaml)
+
+            mock_controller = mock.Mock()
+            q = DistUpgradeQuirks(mock_controller, mock.Mock())
+            q._remove_netplan_match_rpi(netplan_dir)
+
+            # make sure the backup exists and is the same as the original
+            backup_eth0_filename = eth0_config_filename + '.distUpgrade'
+            self.assertTrue(os.path.exists(os.path.join(
+                netplan_dir, backup_eth0_filename)))
+            with open(os.path.join(netplan_dir,
+                      backup_eth0_filename)) as f:
+                self.assertTrue(f.read() == eth0_config_yaml)
+
+            # make sure the modified contents are what we expect
+            with open(os.path.join(netplan_dir, eth0_config_filename)) as f:
+                new_config = yaml.load(f, yaml.SafeLoader)
+                expected_config = yaml.load(expected_eth0_config_yaml,
+                                            yaml.SafeLoader)
+                self.assertTrue(new_config == expected_config)
+
+            # make sure the backup exists and is the same as the original
+            backup_eth1_filename = eth1_config_filename + '.distUpgrade'
+            self.assertTrue(os.path.exists(os.path.join(
+                netplan_dir, backup_eth1_filename)))
+            with open(os.path.join(netplan_dir,
+                      backup_eth1_filename)) as f:
+                self.assertTrue(f.read() == eth1_config_yaml)
+
+            # make sure the modified contents are what we expect
+            with open(os.path.join(netplan_dir, eth1_config_filename)) as f:
+                new_config = yaml.load(f, yaml.SafeLoader)
+                expected_config = yaml.load(expected_eth1_config_yaml,
+                                            yaml.SafeLoader)
+                self.assertTrue(new_config == expected_config)
 
 
 class TestSnapQuirks(unittest.TestCase):
@@ -808,8 +1000,8 @@ class TestSnapQuirks(unittest.TestCase):
         controller = mock.Mock()
         config = mock.Mock()
         q = DistUpgradeQuirks(controller, config)
-        q._from_version = "19.04"
-        q._to_version = "19.10"
+        q._from_version = "20.04"
+        q._to_version = "22.04"
         # Call method under test
 
         controller.cache = {
@@ -857,22 +1049,49 @@ class TestSnapQuirks(unittest.TestCase):
         self.maxDiff = None
         self.assertDictEqual(
             q._snap_list,
-            {'core18': {
+            {'core20': {
                 'channel': 'stable',
                 'command': 'install',
                 'deb': None, 'snap-id': '1234'},
-             'gnome-3-34-1804': {
-                'channel': 'stable/ubuntu-19.10',
+             'gnome-3-38-2004': {
+                'channel': 'stable/ubuntu-22.04',
                 'command': 'install',
                 'deb': None, 'snap-id': '1234'},
-             'gnome-logs': {
-                'command': 'remove'},
              'snap-store': {
-                'channel': 'stable/ubuntu-19.10',
+                'channel': 'stable/ubuntu-22.04',
                 'command': 'install',
                 'deb': 'gnome-software',
                 'snap-id': '1234'}})
 
+    def test_is_deb2snap_metapkg_installed(self):
+        # Prepare the state for testing
+        controller = mock.Mock()
+        config = mock.Mock()
+        q = DistUpgradeQuirks(controller, config)
+        q._from_version = "19.04"
+        q._to_version = "19.10"
+        controller.cache = {
+            'ubuntu-desktop':
+                make_mock_pkg(
+                    name="ubuntu-desktop",
+                    is_installed=True)
+        }
+
+        testdata = [
+            # (input, expected output)
+            ({}, False),
+            ({'metapkg': None}, False),
+            ({'metapkg': 'ubuntu-desktop'}, True),
+            ({'metapkg': 'kubuntu-desktop'}, False),
+            ({'metapkg': ['kubuntu-desktop', 'ubuntu-desktop']}, True),
+            ({'metapkg': ['kubuntu-desktop', 'lubuntu-desktop']}, False),
+        ]
+
+        for data in testdata:
+            self.assertEqual(q._is_deb2snap_metapkg_installed(data[0]),
+                             data[1],
+                             'Expected {1} for input {0}'.format(*data))
+
     @mock.patch("DistUpgrade.DistUpgradeQuirks.get_arch")
     @mock.patch("urllib.request.urlopen")
     def test_calculate_snap_size_requirements(self, urlopen, arch):
@@ -982,6 +1201,37 @@ class TestSnapQuirks(unittest.TestCase):
         # actual deb packages that will have to be removed during the upgrade
         self.assertEqual(controller.forced_obsoletes.append.call_count, 3)
 
+    @mock.patch("builtins.open", new_callable=mock.mock_open,
+                read_data="""\
+processor       : 0
+cpu             : POWER8 (raw), altivec supported
+clock           : 3491.000000MHz
+revision        : 2.0 (pvr 004d 0200)
+
+processor       : 1
+cpu             : POWER8 (raw), altivec supported
+clock           : 3491.000000MHz
+revision        : 2.0 (pvr 004d 0200)
+
+timebase        : 512000000
+platform        : PowerNV
+model           : 8001-22C
+machine         : PowerNV 8001-22C
+firmware        : OPAL
+MMU             : Hash
+""")
+    @mock.patch("DistUpgrade.DistUpgradeQuirks.get_arch",
+                return_value="ppc64el")
+    def test_power8_fail(self, arch, mock_file):
+        controller = mock.Mock()
+        config = mock.Mock()
+        q = DistUpgradeQuirks(controller, config)
+
+        q._test_and_fail_on_power8()
+        self.assertTrue(len(controller.abort.mock_calls))
+
+        mock_file.assert_called_with("/proc/cpuinfo")
+
 
 if __name__ == "__main__":
     import logging
